/*
This is a thread-safe version of malloc and free implemented by me, Nilay Shah. 
It was done for my operating systems course. The underlying data structure is a 
linked list. The linked list keeps track of all spaces in the heap (whether they
are free or not). A couple of things to note:
    -I've minimized the cost of the sbrk() call by using a rather large number
        to extend the heap by (initially and everytime that it's needed)
    -Adjacent memory locations will be coallesced to (1) save time while
        iterating and (2) save overhead space (24 bytes to be exact)
    -This only works for 64-bit machines (hence the 8 byte padding)
*/

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>
#include <stdio.h>

#define SYSTEM_MALLOC 0

pthread_mutex_t region_mutex = PTHREAD_MUTEX_INITIALIZER;

struct node{ //sizeof(struct allNodes) = 24
    void *memLoc;
    unsigned int size;
    int freed; //1 if freed, 0 if being used
    struct node *next;
};


struct node *head = NULL;

/* iterateMemory: Iterates through the linked list structure.*/
void iterateMemory(){
    struct node *temp;
    temp = head;

    while(temp != NULL){
        temp = temp->next;
    }
}

/* coalesceMemory: Coalesces memory when freed.*/
void coalesceMemory(struct node *prev, struct node *cur, struct node *nex){
    if(prev != NULL && prev->freed){
        prev->size += cur->size;
        prev->next = cur->next;
    }  if(nex->freed){
        cur->size += nex->size;
        cur->next = nex->next;
    }
}

/* initMemory: Initializes the memory by calling sbrk().*/
void initMemory(int size){ //10096
    head = sbrk(size);
    head->memLoc = (void *)head + sizeof(struct node);
    head->size = size;// - sizeof(struct node);
    head->freed = 1;
    head->next = NULL;
}

/* baseCase: This is called in order to initialzie the first time this function 
is called.*/
void *baseCase(struct node *new, int size, int sizeWithNode){
    initMemory(4096*33);

    //Make the node the head
    new = head;

    //Setting new to point to the new header
    new += 1;
    new = (void *)new + size;

    //Setting the writable memory location underneath the new header
    new->memLoc += 2; //double check this part!
    new->memLoc = (void *)new + size;

    //Changing the size up
    new->size = head->size - sizeWithNode;
    head->size = sizeWithNode;

    //Making sure the LL is connected properly
    new->next = head->next; //NULL in this case
    head->next = new;

    //Setting the correct nodes to free or not
    new->freed = 1;
    head->freed = 0;

    //Returning the writeable memory location
    pthread_mutex_unlock(&region_mutex);
    return head->memLoc;
}


/* mymalloc: allocates memory on the heap of the requested size. The block
             of memory returned should always be padded so that it begins
             and ends on a word boundary.
     unsigned int size: the number of bytes to allocate.
     retval: a pointer to the block of memory allocated or NULL if the 
             memory could not be allocated. 
             (NOTE: the system also sets errno, but we are not the system, 
                    so you are not required to do so.
*/
void *mymalloc(unsigned int size){
    #if SYSTEM_MALLOC
    return malloc(size);
    #endif

    pthread_mutex_lock(&region_mutex);
    unsigned int sizeWithNode;

    //Add padding
    if(size%8 != 0){
        size += 8-(size%8);
    }

    sizeWithNode = size + sizeof(struct node);
    struct node *new = NULL;

    //If this is the first time you call mymalloc, start off the heap
    if(head == NULL){
        return baseCase(new, size, sizeWithNode);
    }

    struct node *temp, *prev;
    temp = head;

    //Iterate the whole list
    while(temp != NULL){

        //Check if the node is freed
        if(temp->freed){

            //If the node is exactly how much we need
            if(temp->size == sizeWithNode){
                temp->freed = 0;

                pthread_mutex_unlock(&region_mutex);
                return temp->memLoc;

            //If the node is bigger than how much we need..
            } else if(temp->size > sizeWithNode){
                
                //Check if there's enough room to split for a new node, if not, 
                //do this..
                if(temp->size - sizeWithNode < 32){
                    temp->size = sizeWithNode;
                    //temp->next = new;
                    temp->freed = 0;

                    pthread_mutex_unlock(&region_mutex);
                    return temp->memLoc;
                } 

                //There's room to split for a new node, so we can start using 
                //this chunk
                else{
                    //This is the same as what was done with the base case but 
                    //with temp
                    new = temp;
                    new += 1;
                    new = (void *)new + size;

                    new->memLoc += 2;
                    new->memLoc = (void *)new + size;

                    new->size = temp->size - sizeWithNode;
                    temp->size = sizeWithNode;

                    new->next = temp->next;
                    temp->next = new;

                    new->freed = 1;
                    temp->freed = 0;

                    pthread_mutex_unlock(&region_mutex);
                    return temp->memLoc;
                }
            }
        }
        prev = temp;
        temp = temp->next;
    }

    //If there are no free nodes to insert into, the heap has to be extended
    temp = sbrk(4095*33);

    temp->memLoc = (void *)temp + sizeof(struct node);
    temp->size = sizeWithNode;
    temp->freed = 0;
    temp->next = NULL;

    //This is the same as what was done with the base case but with temp
    prev->next = temp;
    new = temp;

    new += 1;
    new = (void *)new + size;

    new->memLoc += 2;
    new->memLoc = (void *)new + size;

    new->size = (4095*33) - sizeWithNode;

    new->next = temp->next; //should be NULL in this case
    temp->next = new;

    new->freed = 1;

    pthread_mutex_unlock(&region_mutex);
    return temp->memLoc;

}


/* myfree: unallocates memory that has been allocated with mymalloc.
     void *ptr: pointer to the first byte of a block of memory allocated by 
                mymalloc.
     retval: 0 if the memory was successfully freed and 1 otherwise.
             (NOTE: the system version of free returns no error.)
*/
unsigned int myfree(void *ptr) {
    #if SYSTEM_MALLOC
    free(ptr);
    return 0;
    #endif

    pthread_mutex_lock(&region_mutex);
    struct node *temp, *prev, *nex;
    
    prev = NULL;
    temp = head;
    nex = temp->next;

    while(temp != NULL){

        //If the node has been found..
        if(temp->memLoc == ptr){
            temp->freed = 1;

            //Coalesce memory
            coalesceMemory(prev, temp, nex);

            pthread_mutex_unlock(&region_mutex);
            return 0;
        }
        prev = temp;
        temp = temp->next;
        nex = temp->next;
    }

    pthread_mutex_unlock(&region_mutex);

    //If it cannot find it, return 1
    return 1;
}
